package ch.unige.solidify;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.maven.plugin.logging.Log;
import org.json.JSONObject;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;

import ch.unige.solidify.process.BaseProcess;
import ch.unige.solidify.process.OAuth2SecurityProcess;
import ch.unige.solidify.process.PathAddSuffixOnOperationIdProcess;
import ch.unige.solidify.process.PathGetCollectionProcess;
import ch.unige.solidify.process.PathNormalizationProcess;
import ch.unige.solidify.process.PathRemoveApplicationPrefixOnOperationIdProcess;
import ch.unige.solidify.process.PathRemoveExampleProcess;
import ch.unige.solidify.process.PathUpdateTagsProcess;
import ch.unige.solidify.process.PathWithIdCreationProcess;
import ch.unige.solidify.process.RemoveNullFieldsProcess;
import ch.unige.solidify.process.SchemaArrayTypeProcess;
import ch.unige.solidify.process.SchemaChangeInfoProcess;
import ch.unige.solidify.process.SchemaCollectionAndPageCreationProcess;
import ch.unige.solidify.process.SchemaExtraDocumentationProcess;
import ch.unige.solidify.process.SchemaNameDefinitionProcess;

public class JsonCleaner {
  private JSONObject json;
  private String stringJsonExtraDocumentation;
  private Map<String, String> migrationNameMap;

  public JsonCleaner() {
    migrationNameMap = new HashMap<>();
  }

  public void load(String inputFilePath, String inputExtraDocumentationFilePath)
          throws IOException {
    json = new JSONObject(loadJsonFile(inputFilePath));
    if (inputExtraDocumentationFilePath != null) {
      stringJsonExtraDocumentation = loadJsonFile(inputExtraDocumentationFilePath);
    }
  }

  private String loadJsonFile(String inputFilePath) throws IOException {
    Path jsonFile = Paths.get(inputFilePath);
    return new String(Files.readAllBytes(jsonFile), StandardCharsets.UTF_8);
  }

  public void run(String contextPath, List<String> modules, String securityName, Log log) {
    // ******************************
    // ** !! Order is important !! **
    // ******************************
    // Normalize Paths
    this.log(log, "Normalizing paths");
    new PathNormalizationProcess(contextPath).process(json);
    // Define new schema name with path
    // Change schema name every web services with new name
    // Migrate schemas with new names
    this.log(log, "Defining new schema name");
    new SchemaNameDefinitionProcess(modules, migrationNameMap, stringJsonExtraDocumentation).process(json);
    // Update schema : remove 'oneOf'
    this.log(log, "Updating schemas: removing 'oneOf'");
    new SchemaArrayTypeProcess().process(json);
    // Update schemas with extra info
    this.log(log, "Updating schemas: adding extry info");
    new SchemaExtraDocumentationProcess(stringJsonExtraDocumentation).process(json);
    // Add creation & lastUpdate fields for each schemas
    // Add changInfo info in schemas
    this.log(log, "Adding creation & last update info");
    new SchemaChangeInfoProcess().process(json);
    // Add collection & page info in schemas
    this.log(log, "Adding collection & page info");
    new SchemaCollectionAndPageCreationProcess().process(json);
    // Update operationId
    this.log(log, "Updating 'operationId'");
    new PathRemoveApplicationPrefixOnOperationIdProcess().process(json);
    // Remove examples nodes in responses and requestBody
    this.log(log, "Removing example nodes");
    new PathRemoveExampleProcess().process(json);
    // Add new path with id (for GET, PATCH, DELETE)
    this.log(log, "Adding new paths for GET, PATH & DELETE");
    new PathWithIdCreationProcess().process(json);
    // Add to existing path with POST verb the other verb GET that return collection
    this.log(log, "Updating paths with POST/GET");
    new PathGetCollectionProcess().process(json);
    // Add suffix (/get, /post, /get-by-id, /delete-by-id, /patch-by-id) to operation id for each
    // verb
    this.log(log, "Adding suffixes");
    new PathAddSuffixOnOperationIdProcess().process(json);
    // Update tags for each web services
    this.log(log, "Updating tags");
    new PathUpdateTagsProcess().process(json);
    // Remove 'null' fields
    this.log(log, "Removing 'null' fields");
    new RemoveNullFieldsProcess().process(json);
    // Add security info for OAuth2
    this.log(log, "Adding security");
    new OAuth2SecurityProcess(securityName).process(json);
  }

  private void log(Log log, String message) {
    if (log == null) {
      System.out.println(message);
    } else {
      log.info(message);
    }
  }

  public void generate(String outputFilePath) throws IOException {
    createDirectory(outputFilePath);
    Path jsonFile = Paths.get(outputFilePath);
    // Use try-with-resource to get auto-closeable writer instance
    try (BufferedWriter writer = Files.newBufferedWriter(jsonFile)) {
      writer.write(this.getSortedJSON());
    }
  }

  private String getSortedJSON() throws JsonParseException, JsonMappingException, IOException {
    String jsonString = this.json.toString(2);
    ObjectMapper mapper = new ObjectMapper().copy()
            .configure(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS, true)
            .configure(MapperFeature.SORT_PROPERTIES_ALPHABETICALLY, true)
            .configure(SerializationFeature.INDENT_OUTPUT, true);
    Object sortedJson = mapper.readValue(jsonString, Object.class);
    return mapper.writeValueAsString(sortedJson);
  }

  private void createDirectory(String outputFilePath) {
    File directory = new File(getDirectoryPath(outputFilePath));
    if (!directory.exists()) {
      directory.mkdirs();
    }
  }

  private String getDirectoryPath(String outputFilePath) {
    if (outputFilePath.endsWith(BaseProcess.URL_SEPARATOR))
      return outputFilePath;
    int lastIndexOfSeparator = outputFilePath.lastIndexOf(BaseProcess.URL_SEPARATOR);
    return outputFilePath.substring(0, lastIndexOfSeparator);
  }

  public JSONObject getJson() {
    return json;
  }

  public void setJson(JSONObject json) {
    this.json = json;
  }

  public String getStringJsonExtraDocumentation() {
    return stringJsonExtraDocumentation;
  }

  public void setStringJsonExtraDocumentation(String stringJsonExtraDocumentation) {
    this.stringJsonExtraDocumentation = stringJsonExtraDocumentation;
  }
}
